@@ -4,9 +4,8 @@
 #include <fcntl.h>
 #include <gif_lib.h>
 
-char
-load(ImlibImage * im, ImlibProgressFunction progress, char progress_granularity,
-     char immediate_load)
+int
+load2(ImlibImage * im, int load_data)
 {
    static const int    intoffset[] = { 0, 4, 2, 1 };
    static const int    intjump[] = { 8, 8, 4, 2 };
@@ -17,18 +16,11 @@ load(ImlibImage * im, ImlibProgressFunction progress, char progress_granularity,
    GifRecordType       rec;
    ColorMapObject     *cmap;
    int                 i, j, done, bg, r, g, b, w = 0, h = 0;
-   float               per = 0.0, per_inc;
-   int                 last_per = 0, last_y = 0;
    int                 transp;
    int                 fd;
+   DATA32              colormap[256];
 
-   done = 0;
-   rows = NULL;
-   transp = -1;
-
-   fd = open(im->real_file, O_RDONLY);
-   if (fd < 0)
-      return 0;
+   fd = dup(fileno(im->fp));
 
 #if GIFLIB_MAJOR >= 5
    gif = DGifOpenFileHandle(fd, NULL);
@@ -36,12 +28,12 @@ load(ImlibImage * im, ImlibProgressFunction progress, char progress_granularity,
    gif = DGifOpenFileHandle(fd);
 #endif
    if (!gif)
-     {
-        close(fd);
-        return 0;
-     }
+      return LOAD_FAIL;
 
-   rc = 0;                      /* Failure */
+   rc = LOAD_FAIL;
+   done = 0;
+   rows = NULL;
+   transp = -1;
 
    do
      {
@@ -50,6 +42,7 @@ load(ImlibImage * im, ImlibProgressFunction progress, char progress_granularity,
              /* PrintGifError(); */
              rec = TERMINATE_RECORD_TYPE;
           }
+
         if ((rec == IMAGE_DESC_RECORD_TYPE) && (!done))
           {
              if (DGifGetImageDesc(gif) == GIF_ERROR)
@@ -58,14 +51,15 @@ load(ImlibImage * im, ImlibProgressFunction progress, char progress_granularity,
                   rec = TERMINATE_RECORD_TYPE;
                   break;
                }
-             w = gif->Image.Width;
-             h = gif->Image.Height;
+
+             im->w = w = gif->Image.Width;
+             im->h = h = gif->Image.Height;
              if (!IMAGE_DIMENSIONS_OK(w, h))
-                goto quit2;
+                goto quit;
 
              rows = calloc(h, sizeof(GifRowType *));
              if (!rows)
-                goto quit2;
+                goto quit;
 
              for (i = 0; i < h; i++)
                {
@@ -114,87 +108,74 @@ load(ImlibImage * im, ImlibProgressFunction progress, char progress_granularity,
    while (rec != TERMINATE_RECORD_TYPE);
 
    if (transp >= 0)
-     {
-        SET_FLAG(im->flags, F_HAS_ALPHA);
-     }
+      SET_FLAG(im->flags, F_HAS_ALPHA);
    else
-     {
-        UNSET_FLAG(im->flags, F_HAS_ALPHA);
-     }
+      UNSET_FLAG(im->flags, F_HAS_ALPHA);
+
    if (!rows)
+      goto quit;
+
+   if (!load_data)
      {
-        goto quit2;
+        rc = LOAD_SUCCESS;
+        goto quit;
      }
 
-   im->w = w;
-   im->h = h;
+   /* Load data */
 
-   if (im->loader || immediate_load || progress)
+   bg = gif->SBackGroundColor;
+   cmap = (gif->Image.ColorMap ? gif->Image.ColorMap : gif->SColorMap);
+   memset(colormap, 0, sizeof(colormap));
+   if (cmap != NULL)
      {
-        DATA32              colormap[256];
-
-        bg = gif->SBackGroundColor;
-        cmap = (gif->Image.ColorMap ? gif->Image.ColorMap : gif->SColorMap);
-        memset(colormap, 0, sizeof(colormap));
-        if (cmap != NULL)
+        for (i = cmap->ColorCount > 256 ? 256 : cmap->ColorCount; i-- > 0;)
           {
-             for (i = cmap->ColorCount > 256 ? 256 : cmap->ColorCount; i-- > 0;)
-               {
-                  r = cmap->Colors[i].Red;
-                  g = cmap->Colors[i].Green;
-                  b = cmap->Colors[i].Blue;
-                  colormap[i] = PIXEL_ARGB(0xff, r, g, b);
-               }
-             /* if bg > cmap->ColorCount, it is transparent black already */
-             if (transp >= 0 && transp < 256)
-                colormap[transp] = bg >= 0 && bg < 256 ?
-                   colormap[bg] & 0x00ffffff : 0x00000000;
+             r = cmap->Colors[i].Red;
+             g = cmap->Colors[i].Green;
+             b = cmap->Colors[i].Blue;
+             colormap[i] = PIXEL_ARGB(0xff, r, g, b);
           }
-
-        ptr = __imlib_AllocateData(im);
-        if (!ptr)
-           goto quit;
-
-        per_inc = 100.0 / (float)h;
-        for (i = 0; i < h; i++)
-          {
-             for (j = 0; j < w; j++)
-               {
-                  *ptr++ = colormap[rows[i][j]];
-               }
-             per += per_inc;
-             if (progress && (((int)per) != last_per)
-                 && (((int)per) % progress_granularity == 0))
-               {
-                  last_per = (int)per;
-                  if (!(progress(im, (int)per, 0, last_y, w, i)))
-                    {
-                       rc = 2;
-                       goto quit;
-                    }
-                  last_y = i;
-               }
-          }
-
-        if (progress)
-           progress(im, 100, 0, last_y, w, h);
+        /* if bg > cmap->ColorCount, it is transparent black already */
+        if (transp >= 0 && transp < 256)
+           colormap[transp] = bg >= 0 && bg < 256 ?
+              colormap[bg] & 0x00ffffff : 0x00000000;
      }
 
-   rc = 1;                      /* Success */
+   ptr = __imlib_AllocateData(im);
+   if (!ptr)
+      goto quit;
+
+   for (i = 0; i < h; i++)
+     {
+        for (j = 0; j < w; j++)
+          {
+             *ptr++ = colormap[rows[i][j]];
+          }
+
+        if (im->lc && __imlib_LoadProgressRows(im, i, 1))
+          {
+             rc = LOAD_BREAK;
+             goto quit;
+          }
+     }
+
+   rc = LOAD_SUCCESS;
 
  quit:
-   for (i = 0; i < h; i++)
-      free(rows[i]);
-   free(rows);
+   if (rows)
+     {
+        for (i = 0; i < h; i++)
+           free(rows[i]);
+        free(rows);
+     }
 
- quit2:
 #if GIFLIB_MAJOR > 5 || (GIFLIB_MAJOR == 5 && GIFLIB_MINOR >= 1)
    DGifCloseFile(gif, NULL);
 #else
    DGifCloseFile(gif);
 #endif
 
-   if (rc == 0)
+   if (rc <= 0)
       __imlib_FreeData(im);
 
    return rc;
@@ -204,11 +185,6 @@ void
 formats(ImlibLoader * l)
 {
    static const char  *const list_formats[] = { "gif" };
-   int                 i;
-
-   l->num_formats = sizeof(list_formats) / sizeof(char *);
-   l->formats = malloc(sizeof(char *) * l->num_formats);
-
-   for (i = 0; i < l->num_formats; i++)
-      l->formats[i] = strdup(list_formats[i]);
+   __imlib_LoaderSetFormats(l, list_formats,
+                            sizeof(list_formats) / sizeof(char *));
 }
