@@ -69,10 +69,9 @@ freeilbm(ILBM * ilbm)
  * Format FORMsizeILBMtag.size....tag.size....tag.size....
  *------------------------------------------------------------------------------*/
 static int
-loadchunks(char *name, ILBM * ilbm, int full)
+loadchunks(FILE * f, ILBM * ilbm, int full)
 {
    CHUNK              *c;
-   FILE               *f;
    size_t              s;
    long                formsize, pos, z;
    int                 ok, seek;
@@ -80,81 +79,76 @@ loadchunks(char *name, ILBM * ilbm, int full)
 
    ok = 0;
 
-   f = fopen(name, "rb");
-   if (f)
+   s = fread(buf, 1, 12, f);
+   if (s == 12 && !memcmp(buf, "FORM", 4) && !memcmp(buf + 8, "ILBM", 4))
      {
-        s = fread(buf, 1, 12, f);
-        if (s == 12 && !memcmp(buf, "FORM", 4) && !memcmp(buf + 8, "ILBM", 4))
+        memset(ilbm, 0, sizeof(*ilbm));
+        formsize = L2RLONG(buf + 4);
+
+        while (1)
           {
-             memset(ilbm, 0, sizeof(*ilbm));
-             formsize = L2RLONG(buf + 4);
+             pos = ftell(f);
+             if (pos < 0 || pos >= formsize + 8)
+                break;          /* Error or FORM data is finished. */
+             seek = 1;
 
-             while (1)
+             s = fread(buf, 1, 8, f);
+             if (s != 8)
+                break;          /* Error or short file. */
+
+             z = L2RLONG(buf + 4);
+             if (z < 0)
+                break;          /* Corrupt file. */
+
+             c = NULL;
+             if (!memcmp(buf, "BMHD", 4))
+                c = &(ilbm->bmhd);
+             else if (full)
                {
-                  pos = ftell(f);
-                  if (pos < 0 || pos >= formsize + 8)
-                     break;     /* Error or FORM data is finished. */
-                  seek = 1;
-
-                  s = fread(buf, 1, 8, f);
-                  if (s != 8)
-                     break;     /* Error or short file. */
-
-                  z = L2RLONG(buf + 4);
-                  if (z < 0)
-                     break;     /* Corrupt file. */
-
-                  c = NULL;
-                  if (!memcmp(buf, "BMHD", 4))
-                     c = &(ilbm->bmhd);
-                  else if (full)
-                    {
-                       if (!memcmp(buf, "CAMG", 4))
-                          c = &(ilbm->camg);
-                       else if (!memcmp(buf, "CMAP", 4))
-                          c = &(ilbm->cmap);
-                       else if (!memcmp(buf, "CTBL", 4))
-                          c = &(ilbm->ctbl);
-                       else if (!memcmp(buf, "SHAM", 4))
-                          c = &(ilbm->sham);
-                       else if (!memcmp(buf, "BODY", 4))
-                          c = &(ilbm->body);
-                    }
-
-                  if (c && !c->data)
-                    {
-                       c->size = z;
-                       c->data = malloc(c->size);
-                       if (!c->data)
-                          break;        /* Out of memory. */
-
-                       s = fread(c->data, 1, c->size, f);
-                       if (s != (size_t)c->size)
-                          break;        /* Error or short file. */
-
-                       seek = 0;
-                       if (!full)
-                         {      /* Only BMHD required. */
-                            ok = 1;
-                            break;
-                         }
-                    }
-
-                  if (pos + 8 + z >= formsize + 8)
-                     break;     /* This was last chunk. */
-
-                  if (seek && fseek(f, z, SEEK_CUR) != 0)
-                     break;
+                  if (!memcmp(buf, "CAMG", 4))
+                     c = &(ilbm->camg);
+                  else if (!memcmp(buf, "CMAP", 4))
+                     c = &(ilbm->cmap);
+                  else if (!memcmp(buf, "CTBL", 4))
+                     c = &(ilbm->ctbl);
+                  else if (!memcmp(buf, "SHAM", 4))
+                     c = &(ilbm->sham);
+                  else if (!memcmp(buf, "BODY", 4))
+                     c = &(ilbm->body);
                }
 
-             /* File may end strangely, especially if body size is uneven, but it's
-              * ok if we have the chunks we want. !full check is already done. */
-             if (ilbm->bmhd.data && ilbm->body.data)
-                ok = 1;
-             if (!ok)
-                freeilbm(ilbm);
+             if (c && !c->data)
+               {
+                  c->size = z;
+                  c->data = malloc(c->size);
+                  if (!c->data)
+                     break;     /* Out of memory. */
+
+                  s = fread(c->data, 1, c->size, f);
+                  if (s != (size_t)c->size)
+                     break;     /* Error or short file. */
+
+                  seek = 0;
+                  if (!full)
+                    {           /* Only BMHD required. */
+                       ok = 1;
+                       break;
+                    }
+               }
+
+             if (pos + 8 + z >= formsize + 8)
+                break;          /* This was last chunk. */
+
+             if (seek && fseek(f, z, SEEK_CUR) != 0)
+                break;
           }
-        fclose(f);
+
+        /* File may end strangely, especially if body size is uneven, but it's
+         * ok if we have the chunks we want. !full check is already done. */
+        if (ilbm->bmhd.data && ilbm->body.data)
+           ok = 1;
+        if (!ok)
+           freeilbm(ilbm);
      }
 
    return ok;
@@ -442,92 +436,89 @@ deplane(DATA32 * row, int w, ILBM * ilbm, unsigned char *plane[])
 }
 
 /*------------------------------------------------------------------------------
- * Loads an image. If im->loader is non-zero, or immediate_load is non-zero, or
- * progress is non-zero, then the file is fully loaded, otherwise only the width
- * and height are read.
+ * Loads an image. If load_data is non-zero then the file is fully loaded,
+ * otherwise only the width and height are read.
  *
  * Imlib2 doesn't support reading comment chunks like ANNO.
  *------------------------------------------------------------------------------*/
-char
-load(ImlibImage * im, ImlibProgressFunction progress, char progress_granularity,
-     char immediate_load)
+int
+load2(ImlibImage * im, int load_data)
 {
+   int                 rc;
    char               *env;
-   int                 cancel, full, i, n, ok, y, z, gran, nexty, prevy;
+   int                 i, n, y, z;
    unsigned char      *plane[40];
    ILBM                ilbm;
 
   /*----------
-   * Load the chunk(s) we're interested in. If full is not true, then we only
+   * Load the chunk(s) we're interested in. If load_data is not true, then we only
    * want the image size and format.
    *----------*/
-   full = (im->loader || immediate_load || progress);
-   ok = loadchunks(im->real_file, &ilbm, full);
-   if (!ok)
-      return 0;
+   rc = loadchunks(im->fp, &ilbm, load_data);
+   if (rc == 0)
+      return LOAD_FAIL;
 
   /*----------
    * Use and check header.
    *----------*/
-   ok = 0;
-   if (ilbm.bmhd.size >= 20)
+   rc = LOAD_FAIL;
+   plane[0] = NULL;
+
+   if (ilbm.bmhd.size < 20)
+      goto quit;
+
+   im->w = L2RWORD(ilbm.bmhd.data);
+   im->h = L2RWORD(ilbm.bmhd.data + 2);
+   if (!IMAGE_DIMENSIONS_OK(im->w, im->h))
+      goto quit;
+
+   ilbm.depth = ilbm.bmhd.data[8];
+   if (ilbm.depth < 1
+       || (ilbm.depth > 8 && ilbm.depth != 24 && ilbm.depth != 32))
+      goto quit;                /* Only 1 to 8, 24, or 32 planes. */
+
+   ilbm.rle = ilbm.bmhd.data[10];
+   if (ilbm.rle < 0 || ilbm.rle > 1)
+      goto quit;                /* Only NONE or RLE compression. */
+
+   ilbm.mask = ilbm.bmhd.data[9];
+
+   if (ilbm.mask || ilbm.depth == 32)
+      SET_FLAG(im->flags, F_HAS_ALPHA);
+   else
+      UNSET_FLAG(im->flags, F_HAS_ALPHA);
+
+   env = getenv("IMLIB2_LBM_NOMASK");
+   if (env
+       && (!strcmp(env, "true") || !strcmp(env, "1") || !strcmp(env, "yes")
+           || !strcmp(env, "on")))
+      UNSET_FLAG(im->flags, F_HAS_ALPHA);
+
+   if (!load_data)
      {
-        ok = 1;
-
-        im->w = L2RWORD(ilbm.bmhd.data);
-        im->h = L2RWORD(ilbm.bmhd.data + 2);
-        if (!IMAGE_DIMENSIONS_OK(im->w, im->h))
-          {
-             ok = 0;
-          }
-
-        ilbm.depth = ilbm.bmhd.data[8];
-        if (ilbm.depth < 1
-            || (ilbm.depth > 8 && ilbm.depth != 24 && ilbm.depth != 32))
-           ok = 0;              /* Only 1 to 8, 24, or 32 planes. */
-
-        ilbm.rle = ilbm.bmhd.data[10];
-        if (ilbm.rle < 0 || ilbm.rle > 1)
-           ok = 0;              /* Only NONE or RLE compression. */
-
-        ilbm.mask = ilbm.bmhd.data[9];
-
-        if (ilbm.mask || ilbm.depth == 32)
-           SET_FLAG(im->flags, F_HAS_ALPHA);
-        else
-           UNSET_FLAG(im->flags, F_HAS_ALPHA);
-
-        env = getenv("IMLIB2_LBM_NOMASK");
-        if (env
-            && (!strcmp(env, "true") || !strcmp(env, "1") || !strcmp(env, "yes")
-                || !strcmp(env, "on")))
-           UNSET_FLAG(im->flags, F_HAS_ALPHA);
-
-        ilbm.ham = 0;
-        ilbm.hbrite = 0;
-        if (ilbm.depth <= 8)
-          {
-             if (ilbm.camg.size == 4)
-               {
-                  if (ilbm.camg.data[2] & 0x08)
-                     ilbm.ham = 1;
-                  if (ilbm.camg.data[3] & 0x80)
-                     ilbm.hbrite = 1;
-               }
-             else
-               {                /* Only guess at ham and hbrite if CMAP is present. */
-                  if (ilbm.depth == 6 && full && ilbm.cmap.size >= 3 * 16)
-                     ilbm.ham = 1;
-                  if (full && !ilbm.ham && ilbm.depth > 1
-                      && ilbm.cmap.size == 3 * (1 << (ilbm.depth - 1)))
-                     ilbm.hbrite = 1;
-               }
-          }
+        rc = LOAD_SUCCESS;
+        goto quit;
      }
-   if (!full || !ok)
+
+   ilbm.ham = 0;
+   ilbm.hbrite = 0;
+   if (ilbm.depth <= 8)
      {
-        freeilbm(&ilbm);
-        return ok;
+        if (ilbm.camg.size == 4)
+          {
+             if (ilbm.camg.data[2] & 0x08)
+                ilbm.ham = 1;
+             if (ilbm.camg.data[3] & 0x80)
+                ilbm.hbrite = 1;
+          }
+        else
+          {                     /* Only guess at ham and hbrite if CMAP is present. */
+             if (ilbm.depth == 6 && ilbm.cmap.size >= 3 * 16)
+                ilbm.ham = 1;
+             if (!ilbm.ham && ilbm.depth > 1
+                 && ilbm.cmap.size == 3 * (1 << (ilbm.depth - 1)))
+                ilbm.hbrite = 1;
+          }
      }
 
   /*----------
@@ -536,72 +527,54 @@ load(ImlibImage * im, ImlibProgressFunction progress, char progress_granularity,
    * from each plane are interleaved, from top to bottom. The first plane is the
    * 0 bit.
    *----------*/
-   ok = 0;
-   cancel = 0;
-   plane[0] = NULL;
-   gran = nexty = 0;
 
    __imlib_AllocateData(im);
+   if (!im->data)
+      goto quit;
+
    n = ilbm.depth;
    if (ilbm.mask == 1)
       n++;
    plane[0] = malloc(((im->w + 15) / 16) * 2 * n);
-   if (im->data && plane[0])
+   if (!plane[0])
+      goto quit;
+
+   for (i = 1; i < n; i++)
+      plane[i] = plane[i - 1] + ((im->w + 15) / 16) * 2;
+
+   z = ((im->w + 15) / 16) * 2 * n;
+
+   scalecmap(&ilbm);
+
+   for (y = 0; y < im->h; y++)
      {
-        for (i = 1; i < n; i++)
-           plane[i] = plane[i - 1] + ((im->w + 15) / 16) * 2;
+        bodyrow(plane[0], z, &ilbm);
 
-        z = ((im->w + 15) / 16) * 2 * n;
+        deplane(im->data + im->w * y, im->w, &ilbm, plane);
+        ilbm.row++;
 
-        if (progress)
+        if (im->lc && __imlib_LoadProgressRows(im, y, 1))
           {
-             prevy = 0;
-             if (progress_granularity <= 0)
-                progress_granularity = 1;
-             gran = progress_granularity;
-             nexty = ((im->h * gran) / 100);
+             rc = LOAD_BREAK;
+             goto quit;
           }
-
-        scalecmap(&ilbm);
-
-        for (y = 0; y < im->h; y++)
-          {
-             bodyrow(plane[0], z, &ilbm);
-
-             deplane(im->data + im->w * y, im->w, &ilbm, plane);
-             ilbm.row++;
-
-             if (progress && (y >= nexty || y == im->h - 1))
-               {
-                  if (!progress
-                      (im, (char)((100 * (y + 1)) / im->h), 0, prevy, im->w,
-                       y + 1))
-                    {
-                       cancel = 1;
-                       break;
-                    }
-                  prevy = y;
-                  gran += progress_granularity;
-                  nexty = ((im->h * gran) / 100);
-               }
-          }
-
-        ok = !cancel;
      }
 
+   rc = LOAD_SUCCESS;
+
+ quit:
   /*----------
    * We either had a successful decode, the user cancelled, or we couldn't get
    * the memory for im->data or plane[0].
    *----------*/
-   if (!ok)
+   if (rc <= 0)
       __imlib_FreeData(im);
 
-   if (plane[0])
-      free(plane[0]);
+   free(plane[0]);
 
    freeilbm(&ilbm);
 
-   return (cancel) ? 2 : ok;
+   return rc;
 }
 
 /*------------------------------------------------------------------------------
@@ -624,11 +597,6 @@ void
 formats(ImlibLoader * l)
 {
    static const char  *const list_formats[] = { "iff", "ilbm", "lbm" };
-   int                 i;
-
-   l->num_formats = sizeof(list_formats) / sizeof(char *);
-   l->formats = malloc(sizeof(char *) * l->num_formats);
-
-   for (i = 0; i < l->num_formats; i++)
-      l->formats[i] = strdup(list_formats[i]);
+   __imlib_LoaderSetFormats(l, list_formats,
+                            sizeof(list_formats) / sizeof(char *));
 }
