@@ -6,77 +6,68 @@
 /* PNG stuff */
 #define PNG_BYTES_TO_CHECK 4
 
+typedef struct {
+   unsigned char       buf[PNG_BYTES_TO_CHECK];
+   unsigned char     **lines;
+} ImLib_PNG_data;
+
 static void
 comment_free(ImlibImage * im, void *data)
 {
    free(data);
 }
 
-char
-load(ImlibImage * im, ImlibProgressFunction progress,
-     char progress_granularity, char immediate_load)
+int
+load2(ImlibImage * im, int load_data)
 {
+   int                 rc;
    png_uint_32         w32, h32;
    int                 w, h;
    char                hasa;
-   FILE               *f;
    png_structp         png_ptr = NULL;
    png_infop           info_ptr = NULL;
    int                 bit_depth, color_type, interlace_type;
-   unsigned char       buf[PNG_BYTES_TO_CHECK];
-
-   f = fopen(im->real_file, "rb");
-   if (!f)
-      return 0;
+   ImLib_PNG_data      pdata;
+   int                 i;
 
    /* read header */
-   hasa = 0;
+   rc = LOAD_FAIL;
+   pdata.lines = NULL;
+
+   if (fread(pdata.buf, 1, PNG_BYTES_TO_CHECK, im->fp) != PNG_BYTES_TO_CHECK)
+      goto quit;
+
+   if (png_sig_cmp(pdata.buf, 0, PNG_BYTES_TO_CHECK))
+      goto quit;
+
+   rewind(im->fp);
 
-   if (fread(buf, 1, PNG_BYTES_TO_CHECK, f) != PNG_BYTES_TO_CHECK)
-     {
-        fclose(f);
-        return 0;
-     }
-   if (png_sig_cmp(buf, 0, PNG_BYTES_TO_CHECK))
-     {
-        fclose(f);
-        return 0;
-     }
-   rewind(f);
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (!png_ptr)
-     {
-        fclose(f);
-        return 0;
-     }
+      goto quit;
+
    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
-     {
-        png_destroy_read_struct(&png_ptr, NULL, NULL);
-        fclose(f);
-        return 0;
-     }
+      goto quit;
+
    if (setjmp(png_jmpbuf(png_ptr)))
      {
-        im->w = 0;
-        png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
-        fclose(f);
-        return 0;
+        rc = LOAD_FAIL;
+        goto quit;
      }
-   png_init_io(png_ptr, f);
+
+   png_init_io(png_ptr, im->fp);
    png_read_info(png_ptr, info_ptr);
    png_get_IHDR(png_ptr, info_ptr, (png_uint_32 *) (&w32),
                 (png_uint_32 *) (&h32), &bit_depth, &color_type,
                 &interlace_type, NULL, NULL);
    if (!IMAGE_DIMENSIONS_OK(w32, h32))
-     {
-        im->w = 0;
-        png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp) NULL);
-        fclose(f);
-        return 0;
-     }
+      goto quit;
+
    im->w = (int)w32;
    im->h = (int)h32;
+
+   hasa = 0;
    if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
       hasa = 1;
    if (color_type == PNG_COLOR_TYPE_RGB_ALPHA)
@@ -88,34 +79,37 @@ load(ImlibImage * im, ImlibProgressFunction progress,
    else
       UNSET_FLAG(im->flags, F_HAS_ALPHA);
 
-   if (im->loader || immediate_load || progress)
+   if (!load_data)
      {
-        unsigned char     **lines;
-        int                 i;
+        rc = LOAD_SUCCESS;
+        goto quit;
+     }
 
-        w = im->w;
-        h = im->h;
+   /* Load data */
 
-        /* Prep for transformations...  ultimately we want ARGB */
-        /* expand palette -> RGB if necessary */
-        if (color_type == PNG_COLOR_TYPE_PALETTE)
-           png_set_palette_to_rgb(png_ptr);
-        /* expand gray (w/reduced bits) -> 8-bit RGB if necessary */
-        if ((color_type == PNG_COLOR_TYPE_GRAY) ||
-            (color_type == PNG_COLOR_TYPE_GRAY_ALPHA))
-          {
-             png_set_gray_to_rgb(png_ptr);
-             if (bit_depth < 8)
-                png_set_expand_gray_1_2_4_to_8(png_ptr);
-          }
-        /* expand transparency entry -> alpha channel if present */
-        if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
-           png_set_tRNS_to_alpha(png_ptr);
-        /* reduce 16bit color -> 8bit color if necessary */
-        if (bit_depth > 8)
-           png_set_strip_16(png_ptr);
-        /* pack all pixels to byte boundaries */
-        png_set_packing(png_ptr);
+   w = im->w;
+   h = im->h;
+
+   /* Prep for transformations...  ultimately we want ARGB */
+   /* expand palette -> RGB if necessary */
+   if (color_type == PNG_COLOR_TYPE_PALETTE)
+      png_set_palette_to_rgb(png_ptr);
+   /* expand gray (w/reduced bits) -> 8-bit RGB if necessary */
+   if ((color_type == PNG_COLOR_TYPE_GRAY) ||
+       (color_type == PNG_COLOR_TYPE_GRAY_ALPHA))
+     {
+        png_set_gray_to_rgb(png_ptr);
+        if (bit_depth < 8)
+           png_set_expand_gray_1_2_4_to_8(png_ptr);
+     }
+   /* expand transparency entry -> alpha channel if present */
+   if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
+      png_set_tRNS_to_alpha(png_ptr);
+   /* reduce 16bit color -> 8bit color if necessary */
+   if (bit_depth > 8)
+      png_set_strip_16(png_ptr);
+   /* pack all pixels to byte boundaries */
+   png_set_packing(png_ptr);
 
 /* note from raster:                                                         */
 /* thanks to mustapha for helping debug this on PPC Linux remotely by        */
@@ -123,84 +117,57 @@ load(ImlibImage * im, ImlibProgressFunction progress,
 /* what the hell was up with the colors                                      */
 /* now png loading should work on big-endian machines nicely                 */
 #ifdef WORDS_BIGENDIAN
-        png_set_swap_alpha(png_ptr);
-        if (!hasa)
-           png_set_filler(png_ptr, 0xff, PNG_FILLER_BEFORE);
+   png_set_swap_alpha(png_ptr);
+   if (!hasa)
+      png_set_filler(png_ptr, 0xff, PNG_FILLER_BEFORE);
 #else
-        png_set_bgr(png_ptr);
-        if (!hasa)
-           png_set_filler(png_ptr, 0xff, PNG_FILLER_AFTER);
+   png_set_bgr(png_ptr);
+   if (!hasa)
+      png_set_filler(png_ptr, 0xff, PNG_FILLER_AFTER);
 #endif
 
-        if (!__imlib_AllocateData(im))
-          {
-             png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp) NULL);
-             fclose(f);
-             return 0;
-          }
-        lines = (unsigned char **)malloc(h * sizeof(unsigned char *));
+   if (!__imlib_AllocateData(im))
+      goto quit;
 
-        if (!lines)
-          {
-             __imlib_FreeData(im);
-             png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp) NULL);
-             fclose(f);
-             return 0;
-          }
-        for (i = 0; i < h; i++)
-           lines[i] = ((unsigned char *)(im->data)) + (i * w * sizeof(DATA32));
-        if (progress)
-          {
-             int                 y, count, prevy, pass, number_passes, per,
-                nrows = 1;
+   pdata.lines = malloc(h * sizeof(unsigned char *));
+   if (!pdata.lines)
+      goto quit;
 
-             count = 0;
-             number_passes = png_set_interlace_handling(png_ptr);
-             for (pass = 0; pass < number_passes; pass++)
+   for (i = 0; i < h; i++)
+      pdata.lines[i] = ((unsigned char *)(im->data)) + (i * w * sizeof(DATA32));
+
+   if (im->lc)
+     {
+        int                 y, pass, n_passes, nrows = 1;
+
+        n_passes = png_set_interlace_handling(png_ptr);
+        for (pass = 0; pass < n_passes; pass++)
+          {
+             __imlib_LoadProgressSetPass(im, pass, n_passes);
+
+             for (y = 0; y < h; y += nrows)
                {
-                  prevy = 0;
-                  per = 0;
-                  for (y = 0; y < h; y += nrows)
-                    {
-                       png_read_rows(png_ptr, &lines[y], NULL, nrows);
+                  png_read_rows(png_ptr, &pdata.lines[y], NULL, nrows);
 
-                       per = (((pass * h) + y) * 100) / (h * number_passes);
-                       if ((per - count) >= progress_granularity)
-                         {
-                            count = per;
-                            if (!progress(im, per, 0, prevy, w, y - prevy + 1))
-                              {
-                                 free(lines);
-                                 png_read_end(png_ptr, info_ptr);
-                                 png_destroy_read_struct(&png_ptr, &info_ptr,
-                                                         (png_infopp) NULL);
-                                 fclose(f);
-                                 return 2;
-                              }
-                            prevy = y + 1;
-                         }
-                    }
-                  if (!progress(im, per, 0, prevy, w, y - prevy + 1))
+                  if (__imlib_LoadProgressRows(im, y, nrows))
                     {
-                       free(lines);
-                       png_read_end(png_ptr, info_ptr);
-                       png_destroy_read_struct(&png_ptr, &info_ptr,
-                                               (png_infopp) NULL);
-                       fclose(f);
-                       return 2;
+                       rc = LOAD_BREAK;
+                       goto quit1;
                     }
                }
           }
-        else
-           png_read_image(png_ptr, lines);
-        free(lines);
-        png_read_end(png_ptr, info_ptr);
      }
+   else
+     {
+        png_read_image(png_ptr, pdata.lines);
+     }
+
+   rc = LOAD_SUCCESS;
+
 #ifdef PNG_TEXT_SUPPORTED
    {
       png_textp           text;
       int                 num;
-      int                 i;
 
       num = 0;
       png_get_text(png_ptr, info_ptr, &text, &num);
@@ -212,14 +179,22 @@ load(ImlibImage * im, ImlibProgressFunction progress,
         }
    }
 #endif
-   png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp) NULL);
-   fclose(f);
-   return 1;
+
+ quit1:
+   png_read_end(png_ptr, info_ptr);
+ quit:
+   free(pdata.lines);
+   png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
+   if (rc <= 0)
+      __imlib_FreeData(im);
+
+   return rc;
 }
 
 char
 save(ImlibImage * im, ImlibProgressFunction progress, char progress_granularity)
 {
+   int                 rc;
    FILE               *f;
    png_structp         png_ptr;
    png_infop           info_ptr;
@@ -228,34 +203,27 @@ save(ImlibImage * im, ImlibProgressFunction progress, char progress_granularity)
    png_bytep           row_ptr, data;
    png_color_8         sig_bit;
    ImlibImageTag      *tag;
-   int                 quality = 75, compression = 3, num_passes = 1, pass;
-
-   if (!im->data)
-      return 0;
+   int                 quality = 75, compression = 3;
+   int                 pass, n_passes = 1;
 
    f = fopen(im->real_file, "wb");
    if (!f)
-      return 0;
+      return LOAD_FAIL;
+
+   rc = LOAD_FAIL;
+   info_ptr = NULL;
+   data = NULL;
+
    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (!png_ptr)
-     {
-        fclose(f);
-        return 0;
-     }
+      goto quit;
+
    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
-     {
-        fclose(f);
-        png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
-        return 0;
-     }
+      goto quit;
+
    if (setjmp(png_jmpbuf(png_ptr)))
-     {
-        fclose(f);
-        png_destroy_write_struct(&png_ptr, (png_infopp) & info_ptr);
-        png_destroy_info_struct(png_ptr, (png_infopp) & info_ptr);
-        return 0;
-     }
+      goto quit;
 
    /* check whether we should use interlacing */
    interlace = PNG_INTERLACE_NONE;
@@ -266,7 +234,6 @@ save(ImlibImage * im, ImlibProgressFunction progress, char progress_granularity)
 #endif
      }
 
-   data = NULL;
    png_init_io(png_ptr, f);
    if (im->flags & F_HAS_ALPHA)
      {
@@ -330,13 +297,16 @@ save(ImlibImage * im, ImlibProgressFunction progress, char progress_granularity)
    png_set_packing(png_ptr);
 
 #ifdef PNG_WRITE_INTERLACING_SUPPORTED
-   num_passes = png_set_interlace_handling(png_ptr);
+   n_passes = png_set_interlace_handling(png_ptr);
 #endif
 
-   for (pass = 0; pass < num_passes; pass++)
+   for (pass = 0; pass < n_passes; pass++)
      {
         ptr = im->data;
 
+        if (im->lc)
+           __imlib_LoadProgressSetPass(im, pass, n_passes);
+
         for (y = 0; y < im->h; y++)
           {
              if (im->flags & F_HAS_ALPHA)
@@ -355,42 +325,30 @@ save(ImlibImage * im, ImlibProgressFunction progress, char progress_granularity)
                }
              png_write_rows(png_ptr, &row_ptr, 1);
 
-             if (progress)
+             if (im->lc && __imlib_LoadProgressRows(im, y, 1))
                {
-                  char                per;
-                  int                 l, pl = 0, pper = 0;
-
-                  per = 100 * (pass + y / (float)im->h) / num_passes;
-                  if ((per - pper) >= progress_granularity)
-                    {
-                       l = y - pl;
-                       if (!progress(im, per, 0, (y - l), im->w, l))
-                         {
-                            if (data)
-                               free(data);
-                            png_write_end(png_ptr, info_ptr);
-                            png_destroy_write_struct(&png_ptr,
-                                                     (png_infopp) & info_ptr);
-                            png_destroy_info_struct(png_ptr,
-                                                    (png_infopp) & info_ptr);
-                            fclose(f);
-                            return 2;
-                         }
-                       pper = per;
-                       pl = y;
-                    }
+                  rc = LOAD_BREAK;
+                  goto quit;
                }
+
              ptr += im->w;
           }
      }
-   if (data)
-      free(data);
+
+   rc = LOAD_SUCCESS;
+
+ quit:
+   free(data);
    png_write_end(png_ptr, info_ptr);
    png_destroy_write_struct(&png_ptr, (png_infopp) & info_ptr);
-   png_destroy_info_struct(png_ptr, (png_infopp) & info_ptr);
+   if (info_ptr)
+      png_destroy_info_struct(png_ptr, (png_infopp) & info_ptr);
+   if (png_ptr)
+      png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
 
    fclose(f);
-   return 1;
+
+   return rc;
 }
 
 /* fills the ImlibLoader struct with a string array of format file */
@@ -404,11 +362,6 @@ void
 formats(ImlibLoader * l)
 {
    static const char  *const list_formats[] = { "png" };
-   int                 i;
-
-   l->num_formats = sizeof(list_formats) / sizeof(char *);
-   l->formats = malloc(sizeof(char *) * l->num_formats);
-
-   for (i = 0; i < l->num_formats; i++)
-      l->formats[i] = strdup(list_formats[i]);
+   __imlib_LoaderSetFormats(l, list_formats,
+                            sizeof(list_formats) / sizeof(char *));
 }
